from collections import defaultdict
from enum import Enum, auto
from pydantic import validate_call
from typing import Optional, Any

from pyfmto.utilities import logger

__all__ = ['Actions', 'ClientPackage', 'DataArchive', 'SyncDataManager']


class Actions(Enum):
    REGISTER = auto()
    QUIT = auto()


class ClientPackage:
    def __init__(self, cid: Optional[int], action: Any):
        self.cid = cid
        self.action = action


class SyncDataManager:
    def __init__(self):
        self._source: dict[int, dict[int, Any]] = defaultdict(dict)
        self._result: dict[int, dict[int, Any]] = defaultdict(dict)

    @validate_call
    def update_src(self, cid: int, version: int, data: Any):
        self._source[cid][version] = data

    @validate_call
    def update_res(self, cid: int, version: int, data: Any):
        self._result[cid][version] = data

    @validate_call
    def lts_src_ver(self, cid: int) -> int:
        data = self._source.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def lts_res_ver(self, cid: int) -> int:
        data = self._result.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def get_src(self, cid: int, version: int):
        if cid not in self._source:
            logger.debug(f"cid={cid} not found in source")
            return None
        data = self._source[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in source")
        return data

    @validate_call
    def get_res(self, cid: int, version: int):
        if cid not in self._result:
            logger.debug(f"cid={cid} not found in result")
            return None
        data = self._result[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in result")
        return data

    @property
    def available_src_ver(self) -> int:
        try:
            return min([max(data.keys()) for data in self._source.values()])
        except ValueError:
            return -1

    @property
    def num_clients(self) -> int:
        return len(self._source)


class DataArchive:
    def __init__(self):
        self.src_data = []
        self.res_data = []

    @property
    def num_src(self) -> int:
        return len(self.src_data)

    @property
    def num_res(self) -> int:
        return len(self.res_data)

    def add_src(self, src_data):
        self.src_data.append(src_data)

    def add_res(self, agg_data):
        self.res_data.append(agg_data)

    def get_latest_res(self):
        return self.res_data[-1] if self.num_res > 0 else None

    def get_latest_src(self):
        return self.src_data[-1] if self.num_src > 0 else None



import pickle
import requests  # type: ignore
import time
import traceback
import wrapt
from abc import abstractmethod, ABC
from collections import defaultdict
from numpy import ndarray
from requests.exceptions import ConnectionError  # type: ignore
from tqdm import tqdm
from typing import final, Any
from yaml import safe_load

from .packages import ClientPackage, Actions
from pyfmto.problems import SingleTaskProblem
from pyfmto.utilities import logger, update_kwargs, titled_tabulate, tabulate_formats as tf, colored

__all__ = [
    'Client',
    'record_runtime'
]


def record_runtime(name=None):
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        start_time = time.time()
        result = wrapped(*args, **kwargs)
        end_time = time.time()
        runtime = end_time - start_time
        if name is None:
            n = wrapped.__name__
        else:
            n = name
        instance.record_round_info(n, f"[{runtime:.2f}s]")
        return result

    return wrapper


class Client(ABC):
    def __init__(self, problem: SingleTaskProblem, **kwargs):
        self._url: str = ''
        self._conn_retry: int = -1
        self.problem = problem
        self._round_info: dict[str, list[str]] = defaultdict(list)

        self.set_addr()

    def set_addr(self, ip='localhost', port=18510, conn_retry: int = 3):
        """
        Configure the server address and port that this client will connect to.

        Parameters:
        -----------
        ip : str, optional
            The IP address of the server. Default is 'localhost'.

        port : int, optional
            The port number on which the server is listening. Default is 18510.

        conn_retry : int, optional
            Maximum number of connection retry attempts if the initial request fails.
            This value determines how many times the client will attempt to reconnect
            before raising an exception. Default is 3.

        Returns:
        --------
        None
            This method does not return any value.
        """
        # noinspection HttpUrlsUsage
        self._url = f"http://{ip}:{port}"
        self._conn_retry = conn_retry

    def __logging_start_info(self):
        param_dict = defaultdict(list)
        param_dict['TaskName'].append(self.problem.name)
        param_dict['Dim'].append(str(self.problem.dim))
        param_dict['Obj'].append(str(self.problem.obj))
        param_dict['IID'].append(str(self.problem.npd))
        param_dict['IniFE'].append(str(self.problem.fe_init))
        param_dict['MaxFE'].append(str(self.problem.fe_max))
        tab = titled_tabulate(
            f"{self.name} Init information", '=',
            param_dict, headers='keys', tablefmt=tf.rounded_grid
        )
        logger.debug(tab)

    def __logging_rounds_info(self):
        if self._round_info:
            tab = titled_tabulate(
                f"{self.name} {self.problem.name}({self.dim}D)",
                '=', self._round_info, headers='keys', tablefmt=tf.rounded_grid
            )
            logger.info(tab)

    def record_round_info(self, name: str, value: str):
        self._round_info[name].append(value)

    @final
    def start(self):
        try:
            logger.info(f"{self.name} started")
            self.__register_id()
            self.__logging_start_info()

            pbar = tqdm(
                total=self.fe_max,
                initial=self.solutions.num_updated,
                desc=self.name, unit="Round",
                ncols=100, position=self.id, leave=False)

            while self.problem.fe_available > 0:
                self.optimize()
                pbar.update(self.solutions.num_updated)

            self.__logging_rounds_info()
            self.send_quit()
            logger.debug(f"{self.name} exit with available FE = {self.problem.fe_available}")
        except Exception:
            self.send_quit()
            if self.id == 1:
                logger.error(traceback.format_exc())
                print(f"Error occurred while in {colored('self.optimize()', 'red')}, "
                      f"see {colored('out/logs/pyfmto.log', 'green')} for detail.")
            logger.info(f"{self.name} exit with available FE = {self.problem.fe_available}")
            raise
        return self.id, self.solutions

    @abstractmethod
    def optimize(self):
        ...  # pragma: no cover

    def __register_id(self):
        pkg = ClientPackage(self.id, Actions.REGISTER)
        self.request_server(pkg)
        logger.debug(f"{self.name} registered")

    @staticmethod
    def deserialize_pickle(package: Any):
        return pickle.loads(package) if package is not None else None

    def request_server(self, package: ClientPackage,
                       repeat: int = 10, interval: float = 1.,
                       msg=None) -> Any:
        """
        Send a request to the server and wait for a response that satisfies a given condition.

        Parameters
        ----------
        package : Any
            The data package to send to the server.
        repeat : int, optional
            The maximum number of attempts to receive a valid response. Default is 10.
        interval : float, optional
            The time interval (in seconds) between attempts. Default is 1 second. Note
            that even the repeat is 1, the interval still effective.
        msg : str, optional
            Message that output to log after every failed repeat using debug level.

        Returns
        -------
        Any
            Return Any, if all request attempts failure, return None

        Notes
        -----
        This method repeatedly sends a request to the server and waits for a response.
        It will continue to attempt to receive a response up to `repeat` times, with a
        delay of `interval` seconds between each attempt.

        The response is considered acceptable if it passes the `check_pkg` method, which
        determines whether the received response is valid based on personalized criteria.
        If the response does not meet these criteria, the method will perform the next repeat.
        """
        if not isinstance(package, ClientPackage):
            raise ValueError("package should be ClientPackage")
        repeat_max = max(1, repeat)
        request_repeat = 1
        failed_retry = 1
        while request_repeat <= repeat_max:
            if msg:
                logger.debug(f"{self.name} [Request retry {request_repeat}/{repeat_max}] {msg}")
            data = pickle.dumps(package)
            try:
                resp = requests.post(
                    url=f"{self._url}/alg-comm",
                    data=data,
                    headers={"Content-Type": "application/x-pickle"}
                )
            except ConnectionError:
                logger.error(f"{self.name} Connection failed {failed_retry} times.")
                time.sleep(interval)
                failed_retry += 1
                if failed_retry > self._conn_retry:
                    raise ConnectionError(f"{self.name} Connection failed {self._conn_retry} times.")
                continue
            pkg = self.deserialize_pickle(resp.content)
            if pkg is not None and self.check_pkg(pkg):
                return pkg
            else:
                time.sleep(interval)
                request_repeat += 1
        raise ConnectionError(f"{self.name} Requested repeat failed for {repeat_max} times")

    def check_pkg(self, pkg) -> bool:
        """
        Determine whether the response is acceptable by check the specific data within it.

        Parameters
        ----------
        pkg :
            The response received from the server.

        Returns
        -------
        bool
            True if the package is acceptable; otherwise, False.

        Notes
        -----
        This method allow additional validation for the response data. By default, it
        does not perform any specific checks and returns `True` for any non-None response.

        Subclasses can override this method to implement custom validation logic. Refer
        to the `EXAMPLE` algorithm for a detailed implementation.
        """
        return True

    def send_quit(self):
        quit_pkg = ClientPackage(self.id, Actions.QUIT)
        self.request_server(quit_pkg)

    def update_kwargs(self, kwargs: dict):
        _docstr = self.__class__.__doc__
        docstr = {} if not _docstr else safe_load(_docstr)
        return update_kwargs(self.__class__.__name__, docstr, kwargs)

    @property
    def id(self):
        """Same with the problem id"""
        return self.problem.id

    @property
    def name(self):
        return f"Client {self.id:<2}"

    @property
    def dim(self) -> int:
        return self.problem.dim

    @property
    def obj(self) -> int:
        return self.problem.obj

    @property
    def fe_init(self) -> int:
        return self.problem.fe_init

    @property
    def fe_max(self) -> int:
        return self.problem.fe_max

    @property
    def y_max(self) -> float:
        return self.solutions.y_max

    @property
    def y_min(self) -> float:
        return self.solutions.y_min

    @property
    def lb(self) -> ndarray:
        return self.problem.lb

    @property
    def ub(self) -> ndarray:
        return self.problem.ub

    @property
    def solutions(self):
        return self.problem.solutions



import time
from setproctitle import setproctitle

from pyfmto.problems import Solution
from pyfmto.utilities import (
    logger, reset_log, show_in_table, clear_console,
    backup_log_to)
from .utils import LauncherUtils, RunSolutions
from ..utilities.loaders import ExperimentConfig, ConfigLoader

__all__ = ['Launcher']


class Launcher:
    exp_idx: int
    exp: ExperimentConfig

    def __init__(self, conf_file: str = 'config.yaml'):
        reset_log()
        clear_console()
        self.conf = ConfigLoader(conf_file).launcher

        # Runtime data
        self._repeat_id = 0
        self._num_clients = 0

    def run(self):
        self._setup()
        for self.exp_idx, self.exp in enumerate(self.conf.experiments):
            if self.conf.save:
                self.exp.init_root()
                self.exp.backup_params()
            self._repeating()
        self._teardown()

    @staticmethod
    def _setup():
        setproctitle("AlgClients")

    def _repeating(self):
        self._update_repeat_id()
        while not self._finished:
            # Init clients
            problem = self.exp.problem.initialize()
            clt_params = self.exp.algorithm.params.get('client', {})
            clients = [self.exp.algorithm.client(p, **clt_params) for p in problem]
            self._iid_info = problem[0].npd
            self._num_clients = len(clients)
            self._show_progress()

            # Launch algorithm
            srv_params = self.exp.algorithm.params.get('server', {})
            with LauncherUtils.running_server(self.exp.algorithm.server, **srv_params):
                results = LauncherUtils.start_clients(clients)
            self._save_results(results)
            self._backup_log()
            self._update_repeat_id()
            clear_console()
            time.sleep(1)

    def _backup_log(self):
        reset_log()
        if self.conf.backup:
            res_file = self.exp.result_name(self._repeat_id)
            log_dir = res_file.with_name('logs')
            log_name = res_file.with_suffix('.log').name
            backup_log_to(log_dir, log_name)

    def _teardown(self):
        clear_console()
        self.conf.show_summary()

    def _show_progress(self):
        curr_rep = self.conf.repeat * self.exp_idx + self._repeat_id - 1
        total_rep = self.conf.total_repeat
        colored_tab, original_tab = show_in_table(
            running=f"{self.exp_idx+1}/{self.conf.n_exp}",
            repeat=f"{self._repeat_id}/{self.conf.repeat}",
            progress=f"[{curr_rep}/{total_rep}][{100 * curr_rep / total_rep:.2f}%]",
            algorithm=self.exp.algorithm.name,
            problem=self.exp.problem.name,
            iid=self._iid_info,
            clients=self._num_clients,
            save=self.conf.save)
        print(colored_tab)
        logger.info(f"\n{original_tab}")

    def _save_results(self, results: list[tuple[int, Solution]]):
        if self.conf.save:
            run_solutions = RunSolutions()
            for cid, solution in results:
                run_solutions.update(cid, solution)
            run_solutions.to_msgpack(self.exp.result_name(self._repeat_id))

    def _update_repeat_id(self):
        if self.conf.save:
            self._repeat_id = self.exp.num_results + 1
        else:
            self._repeat_id += 1

    @property
    def _finished(self):
        return self._repeat_id > self.conf.repeat



# Reference:
#   https://www.bilibili.com/video/BV1sK4y1x7e1
#   https://www.cnblogs.com/kangshuaibo/p/14700833.html

import logging.config
import logging.handlers
import shutil
from pathlib import Path

__all__ = ['logger', 'reset_log', 'backup_log_to']

LOG_HEAD = r"""
                               ____                __
            ____     __  __   / __/  ____ ___     / /_   ____
           / __ \   / / / /  / /_   / __ `__ \   / __/  / __ \
          / /_/ /  / /_/ /  / __/  / / / / / /  / /_   / /_/ /
         / .___/   \__, /  /_/    /_/ /_/ /_/   \__/   \____/
        /_/       /____/

"""
LOG_PATH = Path('out', 'logs')
LOG_PATH.mkdir(parents=True, exist_ok=True)
LOG_FILE = LOG_PATH / 'pyfmto.log'
LOG_BACKUP = LOG_PATH / 'backup.log'
LOG_CONF = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simpleFormatter': {
            'format': '%(levelname)-8s%(asctime)-22s%(filename)16s->line(%(lineno)s)|%(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'pyfmto_handler': {
            'class': 'logging.FileHandler',
            'level': 'DEBUG',
            'formatter': 'simpleFormatter',
            'filename': str(LOG_FILE)
        }
    },
    'loggers': {
        'pyfmto': {
            'level': 'INFO',
            'handlers': ['pyfmto_handler'],
            'propagate': 0
        }
    }
}


def reset_log():
    LOG_PATH.mkdir(parents=True, exist_ok=True)
    if LOG_FILE.exists():
        shutil.copy(LOG_FILE, LOG_BACKUP)
    with LOG_FILE.open('w', encoding='utf-8') as f:
        f.write(LOG_HEAD)


def backup_log_to(dest_dir: Path, filename: str = ''):
    dest_dir.mkdir(parents=True, exist_ok=True)
    if filename:
        dest_file = dest_dir / filename
    else:
        dest_file = dest_dir / LOG_BACKUP.name
    if LOG_BACKUP.exists() and not dest_file.exists():
        shutil.copy(LOG_BACKUP, dest_file)


logging.config.dictConfig(LOG_CONF)
logger = logging.getLogger('pyfmto')



import os
import platform
import time
import wrapt
from pyfmto.utilities import logger
from tabulate import tabulate
from typing import Literal, Optional

__all__ = [
    'colored',
    'timer',
    'clear_console',
    'show_in_table',
    'update_kwargs',
    'titled_tabulate',
    'tabulate_formats',
    'matched_str_head',
]


class TabulatesFormats:
    plain = 'plain'
    simple = 'simple'
    grid = 'grid'
    simple_grid = 'simple_grid'
    rounded_grid = 'rounded_grid'
    heavy_grid = 'heavy_grid'
    mixed_grid = 'mixed_grid'
    double_grid = 'double_grid'
    fancy_grid = 'fancy_grid'
    outline = 'outline'
    simple_outline = 'simple_outline'
    rounded_outline = 'rounded_outline'
    mixed_outline = 'mixed_outline'
    double_outline = 'double_outline'
    fancy_outline = 'fancy_outline'
    pipe = 'pipe'
    presto = 'presto'
    orgtbl = 'orgtbl'
    rst = 'rst'
    mediawiki = 'mediawiki'
    html = 'html'
    latex = 'latex'
    latex_raw = 'latex_raw'
    latex_booktabs = 'latex_booktabs'
    latex_longtable = 'latex_longtable'


tabulate_formats = TabulatesFormats()


def colored(text: str, color: Literal['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'reset']):
    color_map = {
        'black': '\033[30m',
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'reset': '\033[0m'
    }

    if color not in color_map:
        raise ValueError(f"Unsupported color: {color}")

    return f"{color_map[color]}{text}{color_map['reset']}"


def timer(name: Optional[str] = None, where: Literal['log', 'console', 'both'] = 'log'):
    """
    A decorator that records the execution time of a function and outputs it to log, console, or both.

    This decorator wraps a function and measures its execution time. The result can be output to a logger,
    the console (with optional color formatting), or both. An optional custom name can be provided for
    the function in the output message; otherwise, the function's own name is used.

    Parameters
    ----------
    name :
        Custom name to display for the decorated function in the output message.
        If not provided, the function's name will be used.

    where :
        Specifies where to output the runtime message:

        - 'log': Output only to the logger (`ologger.info`).
        - 'console': Output only to the terminal with green-colored formatting.
        - 'both': Output to both the logger and the console. Default is `'log'`.

    Returns
    -------
    function
        A wrapped version of the original function that includes timing behavior.

    Notes
    -----
    - Elapsed time is measured in seconds with three decimal places precision.
    - This decorator uses the `wrapt` library to ensure compatibility with various function types
      and to preserve metadata (e.g., docstrings and function names).

    Examples
    --------
    Example 1: Basic usage with default options

    >>> @record_runtime()
    ... def example_func():
    ...     import time
    ...     time.sleep(0.5)
    ...
    >>> example_func()
    INFO - =============== example_func cost 0.500 s ===============

    Example 2: Custom name and console output

    >>> @record_runtime(name="CustomTask", where="console")
    ... def task():
    ...     pass
    ...
    >>> task()
    =============== CustomTask cost 0.000 s ===============

    Example 3: Output to both log and console

    >>> @record_runtime(where='both')
    ... def demo():
    ...     time.sleep(0.1)
    ...
    >>> demo()
    INFO - =============== demo cost 0.100 s ===============
    =============== demo cost 0.100 s ===============
    """

    # noinspection PyUnusedLocal
    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        start_time = time.time()
        result = wrapped(*args, **kwargs)
        end_time = time.time()
        runtime = end_time - start_time
        if name is not None:
            _name = name
        else:
            _name = wrapped.__name__
        _log = f"{'=' * 15} {_name} cost {runtime:.3f} s {'=' * 15}"
        _console = colored(_log, 'green')
        if where == 'log':
            logger.info(_log)
        elif where == 'console':
            print(_console)
        else:
            logger.info(_log)
            print(_console)
        return result

    return wrapper


def titled_tabulate(title: str, fill_char: str, *args, **kwargs):
    title = ' ' + title if not title.startswith(' ') else title
    title = title + ' ' if not title.endswith(' ') else title
    tab = tabulate(*args, **kwargs)
    tit = title.center(tab.find('\n'), fill_char)
    return f"\n{tit}\n{tab}"


def show_in_table(**kwargs):
    keys, colored_values, original_values = zip(*map(_mapper, kwargs.items()))
    colored_data = dict(zip(keys, colored_values))
    original_data = dict(zip(keys, original_values))
    alignment = ['center'] * len(kwargs)
    colored_tab = tabulate(colored_data, headers='keys', tablefmt='rounded_grid', colalign=alignment)
    original_tab = tabulate(original_data, headers='keys', tablefmt='rounded_grid', colalign=alignment)
    return colored_tab, original_tab


def _mapper(item):
    k, v = item
    if v is True:
        return k, [colored('yes', 'green')], [str(v)]
    elif v is False:
        return k, [colored('no', 'red')], [str(v)]
    elif isinstance(v, int):
        return k, [colored(str(v), 'magenta')], [str(v)]
    elif isinstance(v, float):
        val_str = f"{v:.3f}"
        return k, [colored(val_str, 'green')], [val_str]
    else:
        return k, [str(v)], [str(v)]


def update_kwargs(name, defaults: dict, updates: dict):
    """
    Update ``defaults``  with values from ``updates``.

    This function takes a set of default parameters and a set of updated parameters,
    merges them (with updates taking precedence), and logs a formatted table showing
    the differences. It's useful for configuration management where you want to track
    what values are being used and how they differ from defaults.

    Parameters
    ----------
    name : str
        The name/title for the parameter update operation, used in logging.
    defaults : dict
        Dictionary containing default parameter values.
    updates : dict
        Dictionary containing updated parameter values that override defaults.

    Returns
    -------
    dict
        A new dictionary containing the merged parameters (``defaults`` updated with values from ``updates``).

    Examples
    --------
    >>> defaults = {'a': 0.1, 'b': 10}
    >>> updates = {'a': 0.05, 'c': 5}
    >>> result = update_kwargs("Training Config", defaults, updates)
    >>> print(result)
    {'a': 0.05, 'b': 10}
    """
    if not defaults and not updates:
        return {}
    _log_diff(name, defaults, updates)
    res = defaults.copy()
    for key, value in updates.items():
        if key in defaults:
            res[key] = value
    return res


def _log_diff(name, defaults: dict, updates: dict):
    table_data = []
    for key in set(defaults.keys()).union(updates.keys()):
        default_val = str(defaults[key]) if key in defaults else '-'
        updates_val = str(updates[key]) if key in updates else '-'
        if key in defaults:
            using_val = updates.get(key, defaults[key])
        else:
            using_val = '-'
        table_data.append([key, default_val, updates_val, str(using_val)])
    table = titled_tabulate(
        name,
        '=',
        table_data,
        headers=["Parameter", "Default", "Updates", "Using"],
        tablefmt="rounded_grid",
        colalign=("left", "center", "center", "center"),
        disable_numparse=True,  # tabulate will try to parse string 'True' and 'False' to floats and cause error
    )
    logger.debug(table)


def clear_console():
    if platform.system() == 'Windows':
        os.system('cls')
    else:
        os.system('clear')


def matched_str_head(s: str, str_list: list[str]) -> str:
    for item in str_list:
        if item.startswith(s):
            return item
    return ''
